namespace BP
{
	class AI : BP::ScriptActor
	{
		float onFireTime; //[Serialize] time until fire caught on fire effects turn off
		kActor@ movePoint; //a custom point to move to. Null if not moving to custom point
		kActor@ attackPoint; //a custom point to attack. Null if not attacking a custom point
		array<int> pathPoints; //list of TIDs to each pathpoint
		array<kActor@> pathPointActors;
		int pathPointIndex; //current path point ai is pathing to
		bool isPathMoving;
		bool pathIgnorePlayer;
		float razorWindDamageScale;
		float explosiveDamageScale;
		bool noCauseFighting; //any ai hit by this actor will not set there override target to this actor
		bool noOverrideTarget; //never sets override target to any actor type
		bool isOverridePreDamage;
		//------------------------------------------------------------------------------------------------------------------------
		AI(kActor @actor)
		{
			super(@actor);
			
			kDictMem@ def = self.Definition();
			if (def !is null)
			{
				def.GetFloat("DefaultRazorWindDamageScale", razorWindDamageScale, 1.0f);
				def.GetFloat("DefaultExplosiveDamageScale", explosiveDamageScale, 1.0f);
			}
			else
			{
				razorWindDamageScale = 1.0f;
				explosiveDamageScale = 1.0f;
			}
			
			noOverrideTarget = false;
			noCauseFighting = false;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSerialize(kDict dict) override
		{
			SERIALIZE_FLOAT(onFireTime);
			SERIALIZE_FLOAT(razorWindDamageScale);
			SERIALIZE_FLOAT(explosiveDamageScale);
			SERIALIZE_BOOL(noCauseFighting);
			SERIALIZE_BOOL(noOverrideTarget);
			
			SERIALIZE_BOOL(isPathMoving);
			SERIALIZE_BOOL(pathIgnorePlayer);
			SERIALIZE_INT(pathPointIndex);			
			SERIALIZEARRAY_INT(pathPoints);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserialize(kDict& in dict) override
		{
			DESERIALIZE_FLOAT(onFireTime);
			DESERIALIZE_FLOAT(razorWindDamageScale);
			DESERIALIZE_FLOAT(explosiveDamageScale);
			DESERIALIZE_BOOL(noCauseFighting);
			DESERIALIZE_BOOL(noOverrideTarget);
			
			DESERIALIZE_BOOL(isPathMoving);
			DESERIALIZE_INT(pathPointIndex);
			DESERIALIZE_BOOL(pathIgnorePlayer);
			DESERIALIZEARRAY_INT(pathPoints);
			
			@movePoint = null;
			@attackPoint = null;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnLevelLoad(kDictMem@ pDict) override
		{
			ScriptActor::BP_OnLevelLoad(pDict);
			
			pDict.GetFloat("RazorWindDamageScale", razorWindDamageScale, razorWindDamageScale);
			pDict.GetFloat("ExplosiveDamageScale", explosiveDamageScale, explosiveDamageScale);
			pDict.GetBool("NoOverrideTarget", noOverrideTarget, false);
			pDict.GetBool("noCauseFighting", noCauseFighting, false);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnStart() override
		{
			//because actors not created on level load default health to 100 set spawned enemies to definition start health
			if (!levelLoaded)
			{
				kDictMem@ def = self.Definition();
				if (def !is null)
				{
					float health;
					def.GetFloat("health", health, 1.0f);
					self.Health() = health;
					if (IsMultiplayer())
					{
						float mpHealth;
						if (def.GetFloat("MP_Health", mpHealth))
						{
							self.Health() = mpHealth;
						}
						else
						{
							self.Health() *= 2.0f; //enemies have double health in raptorfest mode
						}
					}
					maxHealth = self.Health();
				}
			}
			
			BP::mapScript.OnAIStarted(@self);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserializeStart() override
		{
			if (isPathMoving)
			{
				SetPath(pathPoints, pathIgnorePlayer, pathPointIndex);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) override
		{
			if (BP::Actor::IsDead(@self))
			{
				return;
			}
			
			float damage = ApplyPreDamageModifiers(dmgInfo, dmgInfo.hits);
			TakePreDamage(dmgInfo, damage);
		}
		//------------------------------------------------------------------------------------------------------------------------
		float ApplyPreDamageModifiers(kDamageInfo& in dmgInfo, float damage)
		{
			if (dmgInfo.particle !is null && dmgInfo.particle.ParticleType() == kParticle_RazorWind)
			{
				damage *= razorWindDamageScale;
			}
			if ((dmgInfo.flags & DF_EXPLOSIVE) != 0)
			{
				damage *= explosiveDamageScale;
			}
			
			if (IsMultiplayer())
			{
				kActor @src = dmgInfo.source;
				if (src !is null && src.Type() == kActor_Player)
				{
					damage *= BP::playerDamage;
				}
			}
				
			return damage;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TakePreDamage(kDamageInfo& in dmgInfo, float damage)
		{
			if (isOverridePreDamage)
			{
				isOverridePreDamage = false;
				//do nothing and take the hit with the custom kDamageInfo
			}
			else
			{
				bool inflictDamageOnSelf = false;
				//Handle Health for normal damage, death flow
				if (self.Health() - damage <= 0.0f)
				{
					self.Health() = dmgInfo.hits;
				}
				else
				{
					if (noOverrideTarget)
					{
						inflictDamageOnSelf = true;
					}
					else
					{
						BP::AI@ srcAIScript = BP::UserEvent::GetAIScript(dmgInfo.source);
						if (srcAIScript !is null && srcAIScript.noCauseFighting)
						{
							inflictDamageOnSelf = true;
						}
						else
						{
							self.Health() += (dmgInfo.hits - damage); //increase health by damage difference
						}
					}
					
					if (inflictDamageOnSelf)
					{
						kDamageInfo damageInfo;
						damageInfo.hits = damage;
						damageInfo.radius = dmgInfo.radius;
						damageInfo.flags = dmgInfo.flags;
						damageInfo.boneFlags = dmgInfo.boneFlags;
						damageInfo.boneID = dmgInfo.boneID;
						@damageInfo.inflictor = @self;
						@damageInfo.source = @self;
						@damageInfo.target = @self;
						@damageInfo.particle = dmgInfo.particle;
						
						isOverridePreDamage = true;						
						if (BUG_DOUBLE_SCRIPT_CALL) ignoreNextOnPreDamage = true; //gets flipped when called by inflictdamage
						self.InflictDamage(damageInfo);
						self.Flags() |= AF_NOINFLICTDAMAGE;
						return;
					}
				}
			}
			
			if (IsMultiplayer())
			{
				if (BP::showPlayerDamage)
				{
					ShowDamage(dmgInfo.source, damage);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDamage(kDamageInfo& in dmgInfo) override
		{
			//Check if damaged with fire effect
			bool hasFlameEffect = false;
			kActor @src = dmgInfo.source;
			if (src !is null)
			{
				kDictMem @def = src.Definition();
				if (def !is null)
				{
					def.GetBool("BP_FlameEffect", hasFlameEffect);
				}
			}
			kParticle @particle = dmgInfo.particle;
			if (particle !is null && particle.ParticleType() == BP_Particle_Flame_Thrower)
			{
				hasFlameEffect = true;
			}
			
			if (hasFlameEffect)
			{
				onFireTime = 4.0f;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeath(kDamageInfo& in dmgInfo) override
		{
			//then ondeath was called because enemy was killed when playing this map before and for some reason death gets called right after onspawn
			if (!IsOnDeathMarked())
			{
				DebugAssert(BP::mapScript !is null, "[BP::AI::BP_OnDeath] mapScript is null");
				BP::mapScript.OnAIDeath(@self, dmgInfo);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnTick() override
		{
			if (isPathMoving)
			{
				if (!BP::Actor::IsDead(@self))
				{
					kActor @sightTarget = @self.EnemyAIComponent().SightTarget().Target();
					if (!pathIgnorePlayer && (sightTarget !is null && sightTarget.Type() == kActor_Player))
					{
						//stop pathing
					}
					else //continue pathing
					{
						BP::Actor::AIIgnorePlayer(@self, pathIgnorePlayer);
						
						kActor @pathPointActor = @pathPointActors[pathPointIndex];
						float radius = pathPointActor.WorldComponent().Radius();
						//if reached path point go to next one
						if (BP::Actor::InRadius(@self, @pathPointActor, radius * radius))
						{
							pathPointIndex++;
							if (pathPointIndex >= int(pathPoints.length()))
							{
								pathPointIndex = 0;
							}
							@pathPointActor = @pathPointActors[pathPointIndex];
							BP::Actor::AIMoveToTarget(@self, @pathPointActor);
						}
					}
				}
			}
			
			if (onFireTime > 0.0f)
			{
				onFireTime -= GAME_DELTA_TIME;
				self.RenderMeshComponent().ToggleHotPoint(1, true);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void MoveToPoint(const kVec3&in point)
		{
			if (BP::Actor::Exists(@movePoint))
			{
				movePoint.WorldComponent().SetNearestPosition(point);
			}
			else
			{
				@movePoint = BP::Spawn::Actor(BP_Actor_WayPoint, point);
				movePoint.SetTarget(@self);
			}
			BP::Actor::AIMoveToTarget(@self, @movePoint);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void AttackPoint(const kVec3&in point)
		{
			if (BP::Actor::Exists(@attackPoint))
			{
				attackPoint.WorldComponent().SetNearestPosition(point);
			}
			else
			{
				@attackPoint = BP::Spawn::Actor(BP_Actor_WayPoint, point);
				attackPoint.SetTarget(@self);
			}
			BP::Actor::AIAttackTarget(@self, @attackPoint);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetPath(const array<int>& tagIDs, bool ignorePlayer, const int startPathIndex = 0)
		{
			pathPoints = tagIDs;
			pathPointIndex = startPathIndex;
			pathIgnorePlayer = ignorePlayer;
			isPathMoving = true;
			int pathPointsLength = int(pathPoints.length());
			pathPointActors.resize(pathPointsLength);
			if (pathPointsLength > 0)
			{
				for (int i = 0; i < pathPointsLength; i++)
				{
					@pathPointActors[i] = BP::Actor::FindActorWithTID(pathPoints[i]);
					DebugAssert(pathPointActors[i] !is null, "[BP::AI::SetPath] path point with TID " + pathPoints[i] + " does not exist");
				}
				BP::Actor::AIMoveToTarget(@self, pathPointActors[pathPointIndex]);
			}			
			BP::Actor::AIIgnorePlayer(@self, pathIgnorePlayer);
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
