namespace BP
{
	#define BP_QUEEN_COMBAT1_DIST					500.0
	#define BP_QUEEN_COMBAT1_DAMAGE			8.0
	#define BP_QUEEN_COMBAT1_ANIMHITTIME		0.5
	#define BP_QUEEN_COMBAT1_WAITTIME			1.0


	const float BP_QUEEN_ASS_HEALTH = 350.0f;
	const float BP_QUEEN_GUNS_HEALTH = 200.0f;
	const float BP_QUEEN_HEAD_HEALTH = 500.0f;
	const float BP_QUEEN_ARMS_HEALTH = 200.0f;
	const float BP_QUEEN_DEADPART_HEALTH = 30.0f; //the maximum amount of health all parts will have when dead (makes it easy and quick to kill the corpses parts)

	const float BP_QUEEN_ASS_DMGSCALE = 2.0f;
	const float BP_QUEEN_GUNS_DMGSCALE = 2.0f;
	const float BP_QUEEN_HEAD_DMGSCALE = 1.5f;
	const float BP_QUEEN_ARMS_DMGSCALE = 2.0f;
	const float BP_QUEEN_BODY_DMGSCALE = 1.0f;
	
	enum BPQueenAttachedMeshes
	{
		BP_QUEEN_MODEL_PART_BODY,
		BP_QUEEN_MODEL_PART_ASS,
		BP_QUEEN_MODEL_PART_GUNS,
		BP_QUEEN_MODEL_PART_HEAD,
		BP_QUEEN_MODEL_PART_LITTLE_ARMS,
		BP_TOTAL_QUEEN_MODEL_PARTS
	};

	class Queen : AI
	{
		bool showFlashEffects;
		float combat1WaitTime;
		array<float> modelHealth(BP_TOTAL_QUEEN_MODEL_PARTS);
		array<float> modelMaxHealth(BP_TOTAL_QUEEN_MODEL_PARTS);
		array<float> modelDamageScale(BP_TOTAL_QUEEN_MODEL_PARTS);
		//------------------------------------------------------------------------------------------------------------------------
		Queen(kActor @actor)
		{
			super(@actor);
			
			showFlashEffects = true;
			combat1WaitTime = 0.0f;
			modelDamageScale[BP_QUEEN_MODEL_PART_BODY] = BP_QUEEN_BODY_DMGSCALE;
			modelDamageScale[BP_QUEEN_MODEL_PART_ASS] = BP_QUEEN_ASS_DMGSCALE;
			modelDamageScale[BP_QUEEN_MODEL_PART_GUNS] = BP_QUEEN_GUNS_DMGSCALE;
			modelDamageScale[BP_QUEEN_MODEL_PART_HEAD] = BP_QUEEN_HEAD_DMGSCALE;
			modelDamageScale[BP_QUEEN_MODEL_PART_LITTLE_ARMS] = BP_QUEEN_ARMS_DMGSCALE;
			modelMaxHealth[BP_QUEEN_MODEL_PART_BODY] = 100;
			modelMaxHealth[BP_QUEEN_MODEL_PART_ASS] = BP_QUEEN_ASS_HEALTH;
			modelMaxHealth[BP_QUEEN_MODEL_PART_GUNS] = BP_QUEEN_GUNS_HEALTH;
			modelMaxHealth[BP_QUEEN_MODEL_PART_HEAD] = BP_QUEEN_HEAD_HEALTH;
			modelMaxHealth[BP_QUEEN_MODEL_PART_LITTLE_ARMS] = BP_QUEEN_ARMS_HEALTH;
			for (int i = 0; i < BP_TOTAL_QUEEN_MODEL_PARTS; i++)
			{
				modelHealth[i] = modelMaxHealth[i];
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSerialize(kDict dict)
		{
			AI::BP_OnSerialize(dict);
			
			SERIALIZE_BOOL(showFlashEffects);
			SERIALIZE_FLOAT(combat1WaitTime);
			SERIALIZEARRAY_FLOAT(modelHealth);
			SERIALIZEARRAY_FLOAT(modelMaxHealth);
			SERIALIZEARRAY_FLOAT(modelDamageScale);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserialize(kDict& in dict)
		{
			AI::BP_OnDeserialize(dict);
			
			DESERIALIZE_BOOL(showFlashEffects);
			DESERIALIZE_FLOAT(combat1WaitTime);
			DESERIALIZEARRAY_FLOAT(modelHealth);
			DESERIALIZEARRAY_FLOAT(modelMaxHealth);
			DESERIALIZEARRAY_FLOAT(modelDamageScale);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnLevelLoad(kDictMem@ pDict) override
		{
			AI::BP_OnLevelLoad(pDict);
			
			float dsAss;
			float dsGuns;
			float dsHead;
			float dsArms;
			float dsBody;
			float assHP;
			float gunsHP;
			float headHP;
			float armsHP;
			pDict.GetBool("ShowFlashEffects", showFlashEffects, false);
			pDict.GetFloat("DamageScale_Ass", dsAss, BP_QUEEN_ASS_DMGSCALE);
			pDict.GetFloat("DamageScale_Guns", dsGuns, BP_QUEEN_GUNS_DMGSCALE);
			pDict.GetFloat("DamageScale_Head", dsHead, BP_QUEEN_HEAD_DMGSCALE);
			pDict.GetFloat("DamageScale_Arms", dsArms, BP_QUEEN_ARMS_DMGSCALE);
			pDict.GetFloat("DamageScale_Body", dsBody, BP_QUEEN_BODY_DMGSCALE);
			pDict.GetFloat("Health_Ass", assHP, BP_QUEEN_ASS_HEALTH);
			pDict.GetFloat("Health_Guns", gunsHP, BP_QUEEN_GUNS_HEALTH);
			pDict.GetFloat("Health_Head", headHP, BP_QUEEN_HEAD_HEALTH);
			pDict.GetFloat("Health_Arms", armsHP, BP_QUEEN_ARMS_HEALTH);
			modelDamageScale[BP_QUEEN_MODEL_PART_BODY] = dsBody;
			modelDamageScale[BP_QUEEN_MODEL_PART_ASS] = dsAss;
			modelDamageScale[BP_QUEEN_MODEL_PART_GUNS] = dsGuns;
			modelDamageScale[BP_QUEEN_MODEL_PART_HEAD] = dsHead;
			modelDamageScale[BP_QUEEN_MODEL_PART_LITTLE_ARMS] = dsArms;
			modelMaxHealth[BP_QUEEN_MODEL_PART_ASS] = assHP;
			modelMaxHealth[BP_QUEEN_MODEL_PART_GUNS] = gunsHP;
			modelMaxHealth[BP_QUEEN_MODEL_PART_HEAD] = headHP;
			modelMaxHealth[BP_QUEEN_MODEL_PART_LITTLE_ARMS] = armsHP;
			for (int i = 0; i < BP_TOTAL_QUEEN_MODEL_PARTS; i++)
			{
				modelHealth[i] = modelMaxHealth[i];
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) override
		{
			if (BP::Actor::IsDead(self))
			{
				self.Flags() |= AF_NOINFLICTDAMAGE;
				for (int i = 0; i < BP_TOTAL_QUEEN_MODEL_PARTS; ++i)
				{
					int modelBit = (1 << i);
					if ((dmgInfo.boneFlags & modelBit) != 0)
					{
						DoPartDamage(i, dmgInfo.hits);
					}
				}
				return;
			}
			
			float damage = ApplyPreDamageModifiers(dmgInfo, dmgInfo.hits);
			
			float maxDamageScale = 0.0f;
			//hit no bones or all bones (explosive does this)
			if (dmgInfo.boneFlags == 0)
			{
				//assume hit body
				maxDamageScale = modelDamageScale[BP_QUEEN_MODEL_PART_BODY];
			}
			else if (dmgInfo.boneFlags == 31)
			{
				maxDamageScale = modelDamageScale[BP_QUEEN_MODEL_PART_BODY];
				//StartFx("PrimagenFlashDamage", 30);
			}
			else
			{
				for (int i = 0; i < BP_TOTAL_QUEEN_MODEL_PARTS; ++i)
				{
					int modelBit = (1 << i);
					if ((dmgInfo.boneFlags & modelBit) != 0)
					{
						if (modelDamageScale[i] > maxDamageScale)
						{
							maxDamageScale = modelDamageScale[i];
						}
						DoPartDamage(i, damage);
					}
				}
			}

			damage *= maxDamageScale;
			TakePreDamage(dmgInfo, damage);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeath(kDamageInfo& in dmgInfo) override
		{
			AI::BP_OnDeath(dmgInfo);
			for (int i = 0; i < BP_TOTAL_QUEEN_MODEL_PARTS; ++i)
			{
				modelHealth[i] = MIN(modelHealth[i], BP_QUEEN_DEADPART_HEALTH);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoPartDamage(int part, float damage)
		{
			if (modelHealth[part] <= 0.0f)
			{
				return;
			}
			
			int modelBit = (1 << part);
			if (part != BP_QUEEN_MODEL_PART_BODY)
			{
				modelHealth[part] -= damage;
			}
			
			if (modelHealth[part] <= 0.0f)
			{
				modelHealth[part] = 0.0f;
				kVec3 fxPos = BP::Actor::CenterOrigin(@self, 1.0f);
				self.PlaySound("sounds/shaders/Head Blown Off.ksnd");
				cRenderMesh.SetAttachedMeshVisibility(part - 1, false);
				PartDestroyParticles(fxPos);
			}
			else
			{
				if (showFlashEffects)
				{
					switch (part)
					{
						case BP_QUEEN_MODEL_PART_ASS:
						case BP_QUEEN_MODEL_PART_GUNS:
						case BP_QUEEN_MODEL_PART_HEAD:
						case BP_QUEEN_MODEL_PART_LITTLE_ARMS:
						{
							StartFx("PrimagenFlashDamage", modelBit);
							break;
						}
					}
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PartDestroyParticles(const kVec3&in pos)
		{
			kQuat rotation(self.Pitch(), 0.0f, self.Yaw());
			BP::Spawn::Particle("particles/Hot_Point_Particle_13.particle", @self, pos, rotation);
			BP::Spawn::Particle("particles/Hot_Point_Particle_14.particle", @self, pos, rotation);
			BP::Spawn::Particle("particles/zPlaceholder_12.particle", @self, pos, rotation);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void StartFx(const kStr&in szFxFile, const int modelBits)
		{
			int bits = 0;
			for (int i = 1; i < BP_TOTAL_QUEEN_MODEL_PARTS + 1; ++i)
			{
				if (((1 << i) & modelBits) != 0)
				{
					bits |= (1 << (i-1));
				}
			}
			cRenderMesh.FxFlags() = bits;
			self.RunFxEvent(szFxFile);
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Checks for combat1 animation in order to apply damage to target
		//------------------------------------------------------------------------------------------------------------------------
		void CheckCombat1Radius()
		{
			combat1WaitTime -= GAME_DELTA_TIME;
			
			if (cAnim.IsPlaying(ANIM_GROUND_ATTACK_COMBAT1) && cAnim.PlayTime() >= BP_QUEEN_COMBAT1_ANIMHITTIME && combat1WaitTime <= 0.0f)
			{
				combat1WaitTime = BP_QUEEN_COMBAT1_WAITTIME;
				
				cAi.GetTargetInfo(false, true);
				if (cAi.SightTarget().Target() is null)
				{
					return;
				}
				
				if (cAi.SightTarget().m_fDist <= BP_QUEEN_COMBAT1_DIST)
				{
					kDamageInfo cDamageInfo;
					cDamageInfo.radius = 10*256;
					cDamageInfo.flags = DF_NORMAL;
					@cDamageInfo.inflictor = @self;
					@cDamageInfo.source = @self;
					@cDamageInfo.target = @cAi.SightTarget().Target();
					@cDamageInfo.particle = null;
					cDamageInfo.hits = BP_QUEEN_COMBAT1_DAMAGE;
					cAi.SightTarget().Target().InflictDamage(cDamageInfo);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnTick() override
		{
			AI::BP_OnTick();
			CheckCombat1Radius();
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
