// const int8 DESTRUCTIBLE_FLAG_CHECK_BULLET_IMPACTS               = (1 << 0);
// const int8 DESTRUCTIBLE_FLAG_CHECK_EXPLOSION_IMPACTS            = (1 << 1);
// const int8 DESTRUCTIBLE_FLAG_CHECK_VANISH_AT_END_OF_TRIG_ANIM   = (1 << 2);
// const int8 DESTRUCTIBLE_FLAG_CHECK_AUTO_TRIG_FOR_PLAYER         = (1 << 3);
// const int8 DESTRUCTIBLE_FLAG_CHECK_AUTO_TRIG_FOR_ACTOR          = (1 << 4);
// const int8 DESTRUCTIBLE_FLAG_CHECK_VANISH_AT_END_OF_DEATH_ANIM  = (1 << 5);
// const int8 DESTRUCTIBLE_FLAG_CHECK_HAS_TRIG_DEATH_STATE         = (1 << 6);

// const int8 DESTRUCTIBLE_MSG_START           = 1;
// const int8 DESTRUCTIBLE_MSG_COLLISION_OFF   = 2;
// const int8 DESTRUCTIBLE_MSG_COLLISION_ON    = 3;
// const int8 DESTRUCTIBLE_MSG_INVISIBLE       = 4;
// const int8 DESTRUCTIBLE_MSG_VISIBLE         = 5;

namespace BP
{
	namespace Actors
	{
		class Destructible : ScriptObject
		{
			kActor@     self;
			int16       m_startHealth;
			uint        m_dwFlags;
			int         m_idleAnim;
			int         m_triggerAnim;
			int         m_triggerParticle;
			int         m_triggerSound;
			uint        m_pickupFlags1;
			uint        m_pickupFlags2;
			bool        m_bSpawnMarked;
			bool        m_bLateCall;
			bool mark;
			//------------------------------------------------------------------------------------------------------------------------
			Destructible(kActor@ actor)
			{
				@self = actor;
				m_bLateCall = false;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnLevelLoad(kDictMem@ pDict)
			{
				pDict.GetInt("startHealth", m_startHealth, 5);
				pDict.GetInt("flags", m_dwFlags, 7);
				pDict.GetInt("idleAnim", m_idleAnim, -1);
				pDict.GetInt("triggerAnim", m_triggerAnim, 0);
				pDict.GetInt("triggerParticle", m_triggerParticle, -1);
				pDict.GetInt("triggerSound", m_triggerSound, -1);
				pDict.GetInt("triggerPickupFlags1", m_pickupFlags1, 0);
				pDict.GetInt("triggerPickupFlags2", m_pickupFlags2, 0);
				pDict.GetBool("mark", mark, true);
				
				self.Health() = m_startHealth;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnSerialize(kDict& out dict)
			{
				SERIALIZE(m_startHealth);
				SERIALIZE(m_dwFlags);
				SERIALIZE(m_idleAnim);
				SERIALIZE(m_triggerAnim);
				SERIALIZE(m_triggerParticle);
				SERIALIZE(m_triggerSound);
				SERIALIZE(m_pickupFlags1);
				SERIALIZE(m_pickupFlags2);
				SERIALIZE(m_bSpawnMarked);
				SERIALIZE(m_bLateCall);
				SERIALIZE_BOOL(mark);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnDeserialize(kDict& in dict)
			{
				DESERIALIZE_INT(m_startHealth);
				DESERIALIZE_INT(m_dwFlags);
				DESERIALIZE_INT(m_idleAnim);
				DESERIALIZE_INT(m_triggerAnim);
				DESERIALIZE_INT(m_triggerParticle);
				DESERIALIZE_INT(m_triggerSound);
				DESERIALIZE_INT(m_pickupFlags1);
				DESERIALIZE_INT(m_pickupFlags2);
				DESERIALIZE_BOOL(m_bSpawnMarked);
				DESERIALIZE_BOOL(m_bLateCall);
				DESERIALIZE_BOOL(mark);
				
				self.Health() = m_startHealth;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnSpawn(void)
			{
				// set flag so that they won't get in the way of projectiles
				if(!(self.RenderMeshComponent() is null))
				{
					self.RenderMeshComponent().Flags() |= RMCF_NOEXPANDJOINTS;
				}
				
				if(self.Deserialized())
				{
					return;
				}
			 
				m_bSpawnMarked = false;
				self.AddComponent("kexModeStateComponent", true);
			 
				if(self.ModeStateComponent() is null)
				{
					Sys.Print("Missing mode component on destructible (" + self.Type() + ")!");
					return;
				}
				
				self.ModeStateComponent().AssignModeTable("DestructibleModeTable");
				
				if(self.IsMarked())
				{
					m_bSpawnMarked = true;
					m_bLateCall = true;
					self.Flags() &= ~AF_IMPORTANT;
					self.ModeStateComponent().SetMode(DESTRUCTIBLE_MODE_TRIGGER);
					return;
				}
				
				self.ModeStateComponent().SetMode(DESTRUCTIBLE_MODE_IDLE);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnPreDamage(kDamageInfo& in dmgInfo)
			{
				bool bTakeDamage = false;
				bool bExplosiveDamage = (dmgInfo.flags & DF_EXPLOSIVE) != 0;
				
				// determine if we can take damage or not
				if((m_dwFlags & DESTRUCTIBLE_FLAG_CHECK_BULLET_IMPACTS) != 0 && !bExplosiveDamage)
				{
					bTakeDamage = true;
				}
				if((m_dwFlags & DESTRUCTIBLE_FLAG_CHECK_EXPLOSION_IMPACTS) != 0 && bExplosiveDamage)
				{
					bTakeDamage = true;
				}
				
				if(!bTakeDamage)
				{
					self.Flags() |= AF_NOINFLICTDAMAGE;
				}
				else
				{
					self.RunFxEvent("MotherFlashDamage");
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnDeath(kDamageInfo& in dmgInfo)
			{
				StartDestruct();
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnTick(void)
			{
				if(self.IsMarked() && m_bLateCall)
				{
					self.SignalOn();
					m_bLateCall = false;
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void StartDestruct()
			{
				self.SignalOn();
				if (mark)
				{
					self.Mark(true);
				}
				self.Flags() &= ~AF_IMPORTANT;
				self.ModeStateComponent().SetMode(DESTRUCTIBLE_MODE_TRIGGER);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnTrigger(kActor@ pInstigator, const int msg)
			{
				switch(self.TriggerMessageID())
				{
				case DESTRUCTIBLE_MSG_START:
					StartDestruct();
					break;
					
				case DESTRUCTIBLE_MSG_COLLISION_OFF:
					if(!(self.WorldComponent() is null))
					{
						self.WorldComponent().Flags() |= WCF_NONSOLID;
					}
					break;
					
				case DESTRUCTIBLE_MSG_COLLISION_ON:
					if(!(self.WorldComponent() is null))
					{
						self.WorldComponent().Flags() &= ~WCF_NONSOLID;
					}
					break;
					
				case DESTRUCTIBLE_MSG_INVISIBLE:
					self.Flags() |= AF_HIDDEN;
					self.Mark(true);
					break;
					
				case DESTRUCTIBLE_MSG_VISIBLE:
					self.Flags() &= ~AF_HIDDEN;
					self.Mark(false);
					break;
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			int MODE_IdleAnim(void)
			{
				return m_idleAnim >= 0 ? m_idleAnim : ANIM_DESTRUCTIBLE_IDLE1;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void MODE_TriggerSetup(void)
			{
				if(m_bSpawnMarked)
				{
					self.AnimTrackComponent().SetLastFrame(true);
					if(!(self.WorldComponent() is null))
					{
						self.WorldComponent().Flags() |= WCF_NONSOLID;
					}
				}
				else
				{
					if(m_triggerSound != -1)
					{
						self.PlaySound(m_triggerSound);
					}
					
					if(m_triggerParticle != -1)
					{
						ParticleFactory.Spawn(m_triggerParticle, self, kQuat(self.Pitch(), 0, self.Yaw()), Math::vecZero);
					}
					
					self.Flags() &= ~AF_NOGENERATE;
					self.GenerateActors(m_pickupFlags1, 0);
					self.Flags() &= ~AF_NOGENERATE;
					self.GenerateActors(m_pickupFlags2, 1);
				
					if(self.Health() <= 0)
					{
						if(!(self.WorldComponent() is null))
						{
							self.WorldComponent().Flags() |= WCF_NONSOLID;
						}
					}
				}
				
				if (!self.AnimTrackComponent().CheckAnimID(MODE_TriggerAnim()))
				{
					self.Remove();
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void MODE_TriggerTick(void)
			{
				if(m_bSpawnMarked)
				{
					self.AnimTrackComponent().SetLastFrame(true);
					m_bSpawnMarked = false;
				}
				
				if((m_dwFlags & DESTRUCTIBLE_FLAG_CHECK_VANISH_AT_END_OF_TRIG_ANIM) != 0 &&
					self.AnimTrackComponent().CycleCompleted())
				{
					self.Remove();
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			int MODE_TriggerAnim(void)
			{
				return m_triggerAnim > 0 ? m_triggerAnim : ANIM_DESTRUCTIBLE_GO1;
			}
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
