// const int8 INTERACTIVE_ANIM_INVISIBLE           = (1 << 0);

// const int8 INTERACTIVE_ANIM_MSG_START           = 1;
// const int8 INTERACTIVE_ANIM_MSG_INVISIBLE       = 2;
// const int8 INTERACTIVE_ANIM_MSG_VISIBLE         = 3;
// const int8 INTERACTIVE_ANIM_MSG_CINEMA          = 4;
// const int8 INTERACTIVE_ANIM_MSG_SOUND           = 5;
// const int8 INTERACTIVE_ANIM_MSG_MUSIC           = 6;
// const int8 INTERACTIVE_ANIM_MSG_AMBIENT         = 7;
// const int8 INTERACTIVE_ANIM_MSG_SWAP_MODEL      = 8;

namespace BP
{
	namespace Actors
	{
		class InteractiveAnim : ScriptActor
		{
			uint        m_dwFlags;
			int         m_startAnim;
			int         m_activeAnim;
			bool        isSetup;
			//------------------------------------------------------------------------------------------------------------------------
			InteractiveAnim(kActor@ actor)
			{
				super(@actor);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnSerialize(kDict dict)
			{
				SERIALIZE_INT(m_dwFlags);
				SERIALIZE_INT(m_startAnim);
				SERIALIZE_INT(m_activeAnim);
				SERIALIZE_BOOL(isSetup);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnDeserialize(kDict& in dict)
			{
				DESERIALIZE_INT(m_dwFlags);
				DESERIALIZE_INT(m_startAnim);
				DESERIALIZE_INT(m_activeAnim);
				DESERIALIZE_BOOL(isSetup);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnLevelLoad(kDictMem@ pDict) override
			{
				pDict.GetInt("flags", m_dwFlags, 0);
				pDict.GetInt("startAnim", m_startAnim, 1);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void Setup()
			{
				if (isSetup)
				{
					return;
				}
				isSetup = true;
				// set flag so that they won't get in the way of projectiles
				if(!(self.RenderMeshComponent() is null))
				{
					self.RenderMeshComponent().Flags() |= RMCF_NOEXPANDJOINTS;
				}
				
				self.AddComponent("kexModeStateComponent", true);
				self.ModeStateComponent().AssignModeTable("InitInteractiveAnimModeTable");
				
				if (Game.ActiveMapID() == kLevel_SaveHub)
				{
					self.Mark(false);
				}
				
				self.AnimTrackComponent().MovementClipFlags() = CF_STATIC|CF_IGNOREBLOCKERS;
				
				// do nothing if loaded from a save game
				if (hasDeserialized)
				{
					return;
				}
				
				if (m_startAnim == 0 && self.Definition() !is null)
				{
					self.Definition().GetInt("SpawnAnim", m_startAnim, 1);
				}
				
				m_activeAnim = m_startAnim;
				self.ModeStateComponent().SetMode(INTER_ANIM_ALL_MODE);
				
				if ((m_dwFlags & INTERACTIVE_ANIM_INVISIBLE) != 0)
				{
					self.Flags() |= AF_HIDDEN;
				}
				
				// if(!(self.WorldComponent() is null))
				// {
					// self.WorldComponent().Flags() |= WCF_NONSOLID;
				// }
				
				if (self.IsMarked())
				{
					self.Flags() &= ~AF_HIDDEN;
				}
				
				self.EnableComponent("kexEnemyAIComponent", false);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnStart() override
			{
				Setup();
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnSpawn() override
			{
				Setup();
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnPreDamage(kDamageInfo& in dmgInfo) override
			{
				self.Flags() |= AF_NOINFLICTDAMAGE;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnTrigger(kActor@ pInstigator, const int msg) override
			{
				Setup();
				
				switch(self.TriggerMessageID())
				{
				case INTERACTIVE_ANIM_MSG_START:
					self.Flags() &= ~AF_HIDDEN;
					StartAnim(self.TriggerData());
					break;
					
				case INTERACTIVE_ANIM_MSG_INVISIBLE:
					self.Flags() |= AF_HIDDEN;
					break;
					
				case INTERACTIVE_ANIM_MSG_VISIBLE:
					self.Flags() &= ~AF_HIDDEN;
					break;
					
				case INTERACTIVE_ANIM_MSG_SOUND:
					self.PlaySound(self.TriggerData());
					break;
					
				case INTERACTIVE_ANIM_MSG_MUSIC:
					Game.PlayMusic(self.TriggerData());
					break;
					
				case INTERACTIVE_ANIM_MSG_AMBIENT:
					World.SetGlobalAmbience(self.TriggerData());
					break;
					
				case INTERACTIVE_ANIM_MSG_SWAP_MODEL:
					self.RenderMeshComponent().SwapOutMesh(self.TriggerData());
					break;
					
				case INTERACTIVE_ANIM_MSG_CINEMA:
					self.Flags() &= ~AF_HIDDEN;
					break;
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void StartAnim(const int animID)
			{
				// force the mode state component to play a new animation
				self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
				
				m_activeAnim = animID;
				self.ModeStateComponent().ForgetAnimState();
				self.ModeStateComponent().SetMode(INTER_ANIM_ALL_MODE, true);
			}
			//------------------------------------------------------------------------------------------------------------------------
			int MODE_InterAnim(void)
			{
				return m_activeAnim;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void MODE_InterAnimTick(void)
			{
				if( m_activeAnim >= ANIM_GROUND_ATTACK_COMBAT1 &&
					m_activeAnim <= ANIM_UNDERWATER_COMBO_END5 &&
					self.AnimTrackComponent().CycleCompleted())
				{
					StartAnim(ANIM_GROUND_IDLE1);
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
