//Raptorfest assault map - expected map settings Time Limit - unlimited, Frag Limit - 50
#define MAP_BASE_PLAYER_SPAWNS_TID 		3000
#define MAP_BASE_AI_SPAWNS_TID 			3001
#define MAP_BASE_AI_SPAWNS_TID2			3002
#define MAP_BASE_AI_SPAWNS_TID3			3003
#define MAP_BASE_DOOR1TID				3004
#define MAP_BASE_DOOR2TID				3005

#define MAP_BASE_PHASE1_GOAL			10
#define MAP_BASE_PHASE2_GOAL			20

namespace BP
{
	namespace Maps
	{
		class Base : ScriptMap
		{
			int phase = -1; //progress in map
			int phaseProgress; //anything from enemies killed to objectives met
			int aiSpawnIndex; //(server only)
			int playerSpawnIndex; //(server only)
			array<kActor@> aiSpawns;
			array<kActor@> playerSpawns;
			array<kActor@> dmSpawns;
			//------------------------------------------------------------------------------------------------------------------------
			Base(kActor@ actor)
			{
				super(actor);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnNetSend()
			{
				// NetWorkWriteMap(array<int> = {phase, phaseProgress});
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnNetReceive(array<int>& int32Args, array<int16>& int16Args)
			{
				int newPhase = int32Args[0];
				int progress = int32Args[1];
				SetPhase(newPhase);
				phaseProgress = progress;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void SetPhase(const int newPhase)
			{
				if (phase != newPhase)
				{
					phase = newPhase;
					phaseProgress = 0;
					aiSpawnIndex = 0;
					playerSpawnIndex = 0;
					
					dmSpawns = BP::Actor::FindActorsWithType(kActor_DeathmatchStart);
					
					switch (phase)
					{
						case 0: //spawn at start
						{
							aiSpawns = BP::Actor::FindActorsWithTID(MAP_BASE_AI_SPAWNS_TID);
							break;
						}
						case 1:
						{
							kActor @door = BP::Actor::FindActorWithTID(MAP_BASE_DOOR1TID);
							DebugAssert(door !is null, "[BP::Maps::Base::SetPhase] phase 1 door is null");
							door.ModeStateComponent().SetMode(0);
							door.SignalOn();
							Hud.AddMessage("Gate 1 Open");
							aiSpawns = BP::Actor::FindActorsWithTID(MAP_BASE_AI_SPAWNS_TID2);
							break;
						}
						default:
						{
							kActor @door = BP::Actor::FindActorWithTID(MAP_BASE_DOOR2TID);
							DebugAssert(door !is null, "[BP::Maps::Base::SetPhase] phase 2 door is null");
							door.ModeStateComponent().SetMode(0);
							door.SignalOn();
							Hud.AddMessage("Final Gate Open");
							aiSpawns = BP::Actor::FindActorsWithTID(MAP_BASE_AI_SPAWNS_TID3);
							
							break;
						}
					}
					
					if (phase > 0)
					{
						//ai won't change position/region if it's far away from dmSpawns so move the spawns to my spawn areas
						for (int i = 0; i < int(dmSpawns.length()); i++)
						{
							dmSpawns[i].WorldComponent().SetRegion(aiSpawns[0].WorldComponent().RegionIndex());
							dmSpawns[i].Origin() = aiSpawns[0].Origin();
						}
					}
					
					playerSpawns = BP::Actor::FindActorsWithTID(MAP_BASE_PLAYER_SPAWNS_TID);
					
					DebugAssert(aiSpawns.length() > 0, "[BP::Map::Base::ScriptMap] no ai start spawns");
					DebugAssert(playerSpawns.length() > 0, "[BP::Map::Base::ScriptMap] no player start spawns");
					if (IsNetServer && phase > 0)
					{
						BP_OnNetSend();
					}
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_UserEvent(const int msg, const float x, const float y, const float z, const float f1, const float f2, const float f3) override
			{
				ScriptMap::BP_UserEvent(msg, x, y, z, f1, f2, f3);
				
				//init phase if hasn't been inited yet
				if (phase == -1)
				{
					SetPhase(0);
				}
				
				switch (msg)
				{
					case USEREVENT_MSG_AI_STARTED:
					{
						kActor @aiActor = @BP::UserEvent::argActors[0];
						DebugAssert(BP::Actor::Exists(@aiActor), "[BP::Map::Base::BP_UserEvent] USEREVENT_MSG_AI_STARTED ai actor does not exist");

						aiSpawnIndex = BP::Array::NextIndex(aiSpawns, aiSpawnIndex);
						kActor @spawnActor = aiSpawns[aiSpawnIndex];

						// kVec3 pos;
						// int ri = aiActor.WorldComponent().GetNearPositionAndRegionIndex(spawnActor.Origin(), pos);
						// aiActor.WorldComponent().SetRegion(ri);
						// aiActor.Origin() = pos;
						
						aiActor.WorldComponent().SetRegion(spawnActor.WorldComponent().RegionIndex());
						aiActor.Origin() = spawnActor.Origin();
						break;
					}
					case USEREVENT_MSG_AI_KILLED:
					{
						kActor @aiActor = @BP::UserEvent::argActors[0];
						DebugAssert(BP::Actor::Exists(@aiActor), "[BP::Map::Base::BP_UserEvent] USEREVENT_MSG_AI_KILLED ai actor does not exist");
						//kActor @aiKiller = @BP::UserEvent::argActors[1]; //handle if aiKiller does not exist
						
						phaseProgress++;
						//DebugLog("phase: " + phase + " phaseProgress: " + phaseProgress);
						if (phase == 0 && phaseProgress >= MAP_BASE_PHASE1_GOAL)
						{
							SetPhase(1);
						}
						else if (phase == 1 && phaseProgress >= MAP_BASE_PHASE2_GOAL)
						{
							SetPhase(2);
						}
						break;
					}
					case USEREVENT_MSG_PLAYER_STARTED:
					case USEREVENT_MSG_PLAYER_REVIVED:
					{
						kActor @playerActor = @BP::UserEvent::argActors[0];
						playerSpawnIndex = BP::Array::NextIndex(playerSpawns, playerSpawnIndex);
						kActor @spawnActor = playerSpawns[playerSpawnIndex];
						playerActor.WorldComponent().SetRegion(spawnActor.WorldComponent().RegionIndex());
						playerActor.Origin() = spawnActor.Origin();
						break;
					}
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnStart() override
			{
				ScriptMap::BP_OnStart();
				
				//only initialize the phase to 0 if server else the clients will receive the phase when they join the game
				if (IsNetServer)
				{
					SetPhase(0);
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			// void BP_OnTick() override
			// {
				// ScriptMap::BP_OnTick();
			// }
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
