//Should have OnActorCreated method so can sort all the actor types to not have to go through all of them

namespace BP
{
	namespace Maps
	{
		class ScriptMap : ScriptActor
		{
			//number of players in the map
			int playerCount;
			
			//time (in seconds) on the map
			float time;
			
			//current MP game mode currently in
			int gameModeMP;
			
			array<kActor@> playerActors;
			//------------------------------------------------------------------------------------------------------------------------
			ScriptMap(kActor@ actor)
			{
				super(actor);
				gameModeMP = BP::Cvar::Get("sv_gamemode").Atoi();
				@BP::map = @self;
				@BP::mapScript = @this;
			}
			//------------------------------------------------------------------------------------------------------------------------
			~ScriptMap()
			{
				playerActors.resize(0);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnPlayerCreated(kActor@ actor)
			{
				if (playerActors.findByRef(@actor) < 0)
				{
					playerActors.insertLast(@actor);
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnPlayerRevive(kActor@ actor)
			{
				
			}
			//------------------------------------------------------------------------------------------------------------------------
			// A player joins the match (also when stopped spectating and is now playing)
			//------------------------------------------------------------------------------------------------------------------------
			void OnPlayerJoin()
			{
				//DebugLog("[BP::Maps::ScriptMap::OnPlayerJoin] There are now " + playerCount + " players in the game");
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnAIStarted(kActor@ actor)
			{
			}
			//------------------------------------------------------------------------------------------------------------------------
			void OnAIDeath(kActor@ actor, kDamageInfo& in dmgInfo)
			{
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnSerialize(kDict dict) override
			{
				SERIALIZEKEY_INT("difficulty", BP::Game::difficulty);
				SERIALIZE_FLOAT(time);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnDeserialize(kDict& in dict) override
			{
				DESERIALIZEKEY_INT("difficulty", BP::Game::difficulty);
				DESERIALIZE_FLOAT(time);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnStart()
			{
				
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnTick() override
			{
				time += GAME_DELTA_TIME;
				
				if (IsMultiplayer())
				{
					//Check if player count changes
					int nPlayers = 0;
					for (int i = int(playerActors.length()) - 1; i >= 0; i--)
					{
						kActor@ actor = @playerActors[i];
						if (!BP::Actor::Exists(actor))
						{
							playerActors.removeAt(i);
							continue;
						}
						
						if (actor.Type() == kActor_Player && BP::Actor::IsMPPlayerInGame(@actor))
						{
							nPlayers++;
						}
					}
					
					if (nPlayers > playerCount)
					{
						playerCount = nPlayers;
						OnPlayerJoin();
						
						if (IsNetServer)
						{
							SetAllNetModsSilent(); //update all new players(is all players) modifier settings
							
							array<int> tids;
							FOREACH_ACTOR
							{
								int tid = actor.TID();
								if (tid >= 0 && tids.find(tid) < 0)
								{
									//do net send for all actors to send to the new clients (is all clients)
									BP::ScriptActor@ script = BP::UserEvent::GetScript(actor);
									if (script !is null)
									{
										script.DoNetSend();
									}
									tids.insertLast(tid);
								}
							}
						}
					}
					else if (nPlayers < playerCount)
					{
						playerCount = nPlayers;
					}
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
