namespace BP
{
	namespace Actors
	{
		class MoveablePlatform : BP::ScriptActor
		{
			bool isMCEnabled = true;
			int platformActorID;
			//------------------------------------------------------------------------------------------------------------------------
			MoveablePlatform(kActor@ actor)
			{
				super(actor);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnStart()
			{
				self.WorldComponent().Flags() |= WCF_INVOKE_COLLIDE_CALLBACK;
				self.MovementComponent().ClipFlags() = CF_CLIPEDGES | CF_DROPOFF | CF_CHECKLINKEDBRIDGES | CF_COLLIDECEILING | CF_COLLIDEFLOOR | CF_NOENTERLAVA | CF_NOENTERSWAMP;
				self.MovementComponent().Flags() |= MCF_PROCESS_MOVEMENT;
				
				platformActorID = BP_Actor_Platform_Crate;
				if (self.Definition() !is null)
				{
					self.Definition().GetInt("platformActorID", platformActorID, BP_Actor_Platform_Crate);
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnNetSend()
			{
				kMovementComponent @mc = self.MovementComponent();
				if (mc !is null)
				{
					array<int> int32Args;
					int32Args.insertLast(self.WorldComponent().RegionIndex());
					int32Args.insertLast(Math::FloatToInt(self.Origin().x));
					int32Args.insertLast(Math::FloatToInt(self.Origin().y));
					int32Args.insertLast(Math::FloatToInt(self.Origin().z));
					int32Args.insertLast(Math::FloatToInt(mc.Velocity().x));
					int32Args.insertLast(Math::FloatToInt(mc.Velocity().y));
					int32Args.insertLast(Math::FloatToInt(mc.Velocity().z));
					if (BP::Actor::Exists(self.GetTarget()))
					{
						int32Args.insertLast(1);
					}
					else
					{
						int32Args.insertLast(0);
					}
					// DebugLog("Sending Barrel - TID: " + self.TID() + " RegionIndex: " + self.WorldComponent().RegionIndex() +
					// " Origin X: " + Math::FloatToInt(self.Origin().x) +
					// " Origin Y: " + Math::FloatToInt(self.Origin().y) +
					// " Origin Z: " + Math::FloatToInt(self.Origin().z) +
					// " Vel X: " + Math::FloatToInt(mc.Velocity().x) +
					// " Vel Y: " + Math::FloatToInt(mc.Velocity().y) +
					// " Vel Z: " + Math::FloatToInt(mc.Velocity().z));
					
					NetWorkWriteActorInt32(@self, int32Args);
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnNetReceive(array<int>& int32Args, array<int16>& int16Args)
			{
				kMovementComponent @mc = self.MovementComponent();
				if (mc !is null)
				{
					self.WorldComponent().SetRegion(int32Args[0]);
					self.Origin().x = Math::IntToFloat(int32Args[1]);
					self.Origin().y = Math::IntToFloat(int32Args[2]);
					self.Origin().z = Math::IntToFloat(int32Args[3]);
					mc.Velocity().x = Math::IntToFloat(int32Args[4]);
					mc.Velocity().y = Math::IntToFloat(int32Args[5]);
					mc.Velocity().z = Math::IntToFloat(int32Args[6]);
					bool shouldHaveTarget = int32Args[7] == 1;
					bool hasTarget = BP::Actor::Exists(self.GetTarget());
					if (shouldHaveTarget)
					{
						if (!hasTarget)
						{
							kActor @actor = BP::Spawn::Actor(platformActorID, self.Origin(), self.Yaw());
							self.SetTarget(@actor);
						}
						
						isMCEnabled = false;
						self.EnableComponent("kexMovementComponent", false);
						mc.Flags() &= ~MCF_PROCESS_MOVEMENT;
					}
					else
					{
						if (hasTarget)
						{
							self.GetTarget().Remove();
						}
						self.SetTarget(@BP::Actor::Null);
						isMCEnabled = true;
						self.EnableComponent("kexMovementComponent", true);
						mc.Flags() |= MCF_PROCESS_MOVEMENT;
					}
					
					// DebugLog("Recieve Barrel - TID: " + self.TID() + " RegionIndex: " + int32Args[0] +
					// " Origin X: " + Math::IntToFloat(int32Args[1]) +
					// " Origin Y: " + Math::IntToFloat(int32Args[2]) +
					// " Origin Z: " + Math::IntToFloat(int32Args[3]) +
					// " Vel X: " + Math::IntToFloat(int32Args[4]) +
					// " Vel Y: " + Math::IntToFloat(int32Args[5]) +
					// " Vel Z: " + Math::IntToFloat(int32Args[6]));
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnCollide(kActor@ pCollider)
			{
				kMovementComponent @colliderMC = pCollider.MovementComponent();
				if (colliderMC !is null && pCollider.Type() == kActor_Player)
				{
					kMovementComponent @mc = self.MovementComponent();
					if (mc !is null)
					{
						kVec3 dir = colliderMC.Velocity();
						dir.z = 0.0f;
						mc.Velocity() += dir;
						isMCEnabled = true;
						self.EnableComponent("kexMovementComponent", true);
						if (BP::Actor::Exists(self.GetTarget()))
						{
							self.GetTarget().Remove();
						}
						self.SetTarget(@BP::Actor::Null);
						mc.Flags() |= MCF_PROCESS_MOVEMENT;
						ScheduleNetSync();
					}
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnTick()
			{
				if (isMCEnabled)
				{
					kMovementComponent @mc = self.MovementComponent();
					if (Math::Vec3InRange(mc.Velocity(), 0.01f))
					{
						isMCEnabled = false;
						self.EnableComponent("kexMovementComponent", false);
						kActor @actor = BP::Spawn::Actor(platformActorID, self.Origin(), self.Yaw());
						self.SetTarget(@actor);
					}
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
