namespace BP
{
	namespace Actors
	{
		class MoverBlocker : ScriptActor
		{
			kVec3 movement;
			float moveTime;
			int soundOpen;
			int soundClose;
			
			kVec3 startPos;
			float curMoveTime;
			//------------------------------------------------------------------------------------------------------------------------
			MoverBlocker(kActor @actor)
			{
				super(@actor);
				startPos = self.Origin();
				curMoveTime = 0.0f;
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnLevelLoad(kDictMem@ pDict) override
			{
				pDict.GetVector("movement", movement);
				pDict.GetFloat("moveTime", moveTime);
				pDict.GetInt("soundOpen", soundOpen);
				pDict.GetInt("soundClose", soundClose);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnSerialize(kDict dict) override
			{
				SERIALIZE_VECTOR(movement);
				SERIALIZE_FLOAT(moveTime);
				SERIALIZE_INT(soundOpen);
				SERIALIZE_INT(soundClose);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnDeserialize(kDict& in dict) override
			{
				DESERIALIZE_VECTOR(movement);
				DESERIALIZE_FLOAT(moveTime);
				DESERIALIZE_INT(soundOpen);
				DESERIALIZE_INT(soundClose);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnSpawn() override
			{
				self.Flags() |= AF_NODAMAGE;
				
				if (self.Deserialized())
				{
					return;
				}
				
				if (self.IsMarked())
				{
					Open();
				}
				else
				{
					Close();
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnTick() override
			{
				if (!self.IsMarked())
				{
					curMoveTime = Math::Max(curMoveTime - GAME_DELTA_TIME, 0.0f);
				}
				else
				{
					curMoveTime = Math::Min(curMoveTime + GAME_DELTA_TIME, moveTime);
				}
				
				float lerpTime = curMoveTime / moveTime;
				float mx = Math::Lerp(0.0f, movement.x, lerpTime);
				float my = Math::Lerp(0.0f, movement.y, lerpTime);
				float mz = Math::Lerp(0.0f, movement.z, lerpTime);
				self.Origin() = startPos + kVec3(mx, my, mz);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void Open()
			{
				self.PlaySound(soundOpen);
				Event.RunAction(self, null, ACTION_UNBLOCK, Math::vecZero);
				self.SignalOn(); //might need to be called onStart
				self.Mark(true);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void Close()
			{
				self.PlaySound(soundClose);
				Event.RunAction(self, null, ACTION_BLOCK, Math::vecZero);
				self.SignalOff();
				self.Mark(false);
			}
			//------------------------------------------------------------------------------------------------------------------------
			void BP_OnTrigger(kActor@ pInstigator, const int msg) override
			{
				switch (self.TriggerMessageID())
				{
					case 1: //open
					{
						if (!self.IsMarked())
						{
							Open();
						}
						break;
					}
					case 2: //close
					{
						if (self.IsMarked())
						{
							Close();
						}
						break;
					}
					default:
					{
						break;
					}
				}
			}
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
