namespace BP
{
	#define ACTOR_HB_TYPE_NONE 			0
	#define ACTOR_HB_TYPE_RED1			1
	#define ACTOR_HB_TYPE_BLUE2			2
		
	int lastInstanceID;
	abstract class ScriptActor : ScriptObject
	{
		kActor@ self;
		int instanceID;
		float maxHealth;
		int healthBarType; //type of health bar this actor has
		int healthBarIndex;
		bool hasTicked;
		bool hasDeserialized;
		bool levelLoaded; //is true if this actor was created on level load
		float scheduledNetSyncTime;
		kEnemyAIComponent@ cAi;
		kAnimTrackComponent@ cAnim;
		kWorldComponent@ cWorld;
		kMovementComponent@ cMovement;
		kModeStateComponent@ cModeState;
		kRenderMeshComponent@ cRenderMesh;

		bool ignoreNextOnPreDamage = true;
		bool ignoreNextOnDamage = true;
		bool ignoreNextOnDeath = true;
		bool ignoreNextOnSpawn = true;
		//------------------------------------------------------------------------------------------------------------------------
		ScriptActor(kActor@ actor)
		{
			@self = actor;
			
			lastInstanceID += (lastInstanceID == -1) ? 2 : 1;
			instanceID = lastInstanceID;
			@cAi = self.EnemyAIComponent();
			@cAnim = self.AnimTrackComponent();
			@cWorld = self.WorldComponent();
			@cMovement = self.MovementComponent();
			@cModeState = self.ModeStateComponent();
			@cRenderMesh = self.RenderMeshComponent();
			
			kDictMem@ def = self.Definition();
			if (def !is null)
			{
				if (cWorld !is null && IsMultiplayer())
				{
					float mpRadius;
					float mpWallRadius;
					float mpHeight;
					float mpHeightOffset;
					if (def.GetFloat("MP_Radius", mpRadius))
						cWorld.Radius() = mpRadius;
					if (def.GetFloat("MP_WallRadius", mpWallRadius))
						cWorld.WallRadius() = mpWallRadius;
					if (def.GetFloat("MP_Height", mpHeight))
						cWorld.Height() = mpHeight;
					if (def.GetFloat("MP_HeightOffset", mpHeightOffset))
						cWorld.HeightOffset() = mpHeightOffset;
				}
			}
			
			maxHealth = self.Health();
			healthBarType = ACTOR_HB_TYPE_NONE;
			healthBarIndex = 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		~ScriptActor()
		{
			if (healthBarType != ACTOR_HB_TYPE_NONE)
				HideHealthBar();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnSpawn() final
		{
			//DebugLog("[ScriptActor::OnSpawn] type: " + self.Type());
			if (BUG_DOUBLE_SCRIPT_CALL)
			{
				ignoreNextOnSpawn = !ignoreNextOnSpawn;
				if (ignoreNextOnSpawn)
				{
					return;
				}
			}
			
			levelLoaded = true;
			BP_OnSpawn();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnBeginLevel() final
		{
			BP_OnBeginLevel();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnLevelLoad(kDictMem@ pDict) final
		{
			//DebugLog("[ScriptActor::OnLevelLoad] type: " + self.Type());
			BP_OnLevelLoad(@pDict);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnSerialize(kDict& out dict) final
		{
			SERIALIZE_INT(instanceID);
			SERIALIZE_BOOL(levelLoaded);
			SERIALIZE_FLOAT(maxHealth);
			SERIALIZE_INT(healthBarType);
			BP_OnSerialize(dict);
			if (healthBarType != ACTOR_HB_TYPE_NONE)
			{
				ShowHealthBar(healthBarType);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnDeserialize(kDict& in dict) final
		{			
			DESERIALIZE_INT(instanceID);
			DESERIALIZE_BOOL(levelLoaded);
			DESERIALIZE_FLOAT(maxHealth);
			DESERIALIZE_INT(healthBarType);
			BP_OnDeserialize(dict);
			hasDeserialized = true;
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnTrigger(kActor@ pInstigator, const int msg) final
		{
			switch(self.TriggerMessageID())
			{
				case TRIGGER_MSG_HEALTHBAR:
				{
					if (self.TriggerData() == ACTOR_HB_TYPE_NONE)
					{
						HideHealthBar();
					}
					else
					{
						ShowHealthBar(self.TriggerData());
					}
					break;
				}
			}
			BP_OnTrigger(@pInstigator, msg);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnTick() final
		{
			if (!hasTicked)
			{
				if (self.GameTicks() == 0)
				{
					BP_OnStart();
				}
				if (hasDeserialized)
				{
					@cAi = self.EnemyAIComponent();
					@cAnim = self.AnimTrackComponent();
					@cWorld = self.WorldComponent();
					@cMovement = self.MovementComponent();
					@cModeState = self.ModeStateComponent();
					@cRenderMesh = self.RenderMeshComponent();
			
					BP_OnDeserializeStart();
				}
				hasTicked = true;
			}
			
			if (IsNetServer && scheduledNetSyncTime > 0.0f)
			{
				scheduledNetSyncTime -= GAME_DELTA_TIME;
				if (scheduledNetSyncTime <= 0.0f)
				{
					DoNetSend();
				}
			}
			
			TickHealthBar();
			BP_OnTick();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ScheduleNetSync()
		{
			if (scheduledNetSyncTime <= 0.0f)
			{
				scheduledNetSyncTime = NET_SYNC_TIME;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnEndLevel() final
		{
			BP_OnEndLevel();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnPreDamage(kDamageInfo& in dmgInfo) final
		{
//			DebugLog("[OnPreDamage] Type: " + self.Type() + " ticks: " + self.GameTicks());
			if (BUG_DOUBLE_SCRIPT_CALL)
			{
				ignoreNextOnPreDamage = !ignoreNextOnPreDamage;
				if (ignoreNextOnPreDamage)
				{
					return;
				}
			}
			BP_OnPreDamage(dmgInfo);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnDamage(kDamageInfo& in dmgInfo) final
		{
//			DebugLog("[OnDamage] Type: " + self.Type() + " ticks: " + self.GameTicks());
			
			if (BUG_DOUBLE_SCRIPT_CALL)
			{
				ignoreNextOnDamage = !ignoreNextOnDamage;
				if (ignoreNextOnDamage)
				{
					return;
				}
			}
			BP_OnDamage(dmgInfo);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnDeath(kDamageInfo& in dmgInfo) final
		{
//			DebugLog("[OnDeath] Type: " + self.Type() + " ticks: " + self.GameTicks());
			
			if (BUG_DOUBLE_SCRIPT_CALL)
			{
				ignoreNextOnDeath = !ignoreNextOnDeath;
				if (ignoreNextOnDeath)
				{
					return;
				}
			}
			BP_OnDeath(dmgInfo);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnTouch(kActor@ pInstigator) final
		{
			BP_OnTouch(@pInstigator);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnCollide(kActor@ pCollider) final
		{
			BP_OnCollide(@pCollider);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void UserEvent(const float x, const float y, const float z, const float f1, const float f2, const float f3, const float f4) final
		{
			int msg = int(f1);
			
			switch (msg)
			{
				case USEREVENT_MSG_GET_SCRIPT:
				{
					@BP::UserEvent::scriptResult = @this;
					break;
				}
			}

			BP_UserEvent(msg, x, y, z, f2, f3, f4);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoNetSend()
		{
			scheduledNetSyncTime = 0.0f;
			BP_OnNetSend();
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns true if actor OnDeath is called right after spawn because actor was killed when playing this map before
		// and so OnDeath gets called right after OnSpawn
		//------------------------------------------------------------------------------------------------------------------------
		bool IsOnDeathMarked()
		{
			return !levelLoaded && !hasTicked && self.IsMarked();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ShowDamage(kActor @src, const float damage)
		{
			if (!BP::Actor::Exists(src))
				return;
				
			int damageDisplay = (damage > 0.0f && damage < 1.0f) ? Math::CeilToInt(damage) : int(damage);
			BP::Actor::ShowWorldText(@src, @self, "" + damageDisplay);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void HideHealthBar()
		{
			healthBarType = ACTOR_HB_TYPE_NONE;
			Hud.DisableBar(healthBarIndex);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ShowHealthBar(const int type)
		{
			if (type == ACTOR_HB_TYPE_NONE)
				return;
				
			healthBarType = type;
			switch (healthBarType)
			{
				case ACTOR_HB_TYPE_BLUE2:
					healthBarIndex = 1;
					Hud.SetBar(healthBarIndex, 50, 50, 255, 255, 100);
					break;
				case ACTOR_HB_TYPE_RED1:
				default:
				{
					healthBarIndex = 0;
					Hud.SetBar(healthBarIndex, 255, 50, 50, 255, 100);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TickHealthBar()
		{
			if (BP::Actor::IsDead(self) || BP::Actor::IsHidden(self))
			{
				HideHealthBar();
				return;
			}
			
			if (healthBarType != ACTOR_HB_TYPE_NONE)
			{
				if (maxHealth > 0.0f)
				{
					int healthPercent = int((self.Health() / maxHealth) * 100.0f);
					Hud.UpdateBar(healthBarIndex, MAX(healthPercent, 0));
				}
				else
				{
					Hud.UpdateBar(healthBarIndex, 0);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSpawn() {}
		void BP_OnBeginLevel() {}
		void BP_OnEndLevel() {}
		void BP_OnLevelLoad(kDictMem@ pDict) {}
		void BP_OnSerialize(kDict dict) {}
		void BP_OnDeserialize(kDict& in dict) {}
		void BP_OnTrigger(kActor@ pInstigator, const int msg) {}
		void BP_OnTick() {}
		void BP_OnStart() {} //called once before tick and never again even after loading a save game
		void BP_OnDeserializeStart() {} //called once before tick after loading a save game
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) {}
		void BP_OnDamage(kDamageInfo& in dmgInfo) {}
		void BP_OnDeath(kDamageInfo& in dmgInfo) {}
		void BP_OnTouch(kActor@ pInstigator) {}
		void BP_OnCollide(kActor@ pCollider) {}
		void BP_UserEvent(const int msg, const float x, const float y, const float z, const float f1, const float f2, const float f3) {}
		void BP_OnNetSend() {} // called when need to sync with players
		void BP_OnNetReceive(array<int>& int32Args, array<int16>& int16Args) {} //called when receiving Network data
		//------------------------------------------------------------------------------------------------------------------------
	};
}
