namespace BP
{
	namespace Def
	{
		int maxWeapons;
		array<BP::Def::Weapon> weapons;
		
		void LoadWeapons()
		{
			//load weapon info
			kIndexDefManager iDefManager;
			iDefManager.LoadFile("defs/weaponInfo.txt");
			maxWeapons =  iDefManager.FindHighestIndex() + 1;
			weapons.resize(maxWeapons);
			array<kStr> weaponAmmoNames(maxWeapons);
			array<kStr> weaponAltAmmoNames(maxWeapons);
			
			for (int i = 0; i < maxWeapons; i++)
			{
				kDictMem@ dict = iDefManager.GetEntry(i);
				if (dict !is null)
				{
					kStr name;
					if (dict.GetString("ammo", name))
					{
						weaponAmmoNames[i] = name;
					}
					if (dict.GetString("altAmmo", name))
					{
						weaponAltAmmoNames[i] = name;
					}
					
					Weapon weapon;
					dict.GetInt("actor", weapon.actorID);
					dict.GetInt("pickupTypeID", weapon.pickupActorID);
					dict.GetInt("charMeshIndex", weapon.charMeshIndex);
					dict.GetInt("initialAmmoAmount", weapon.ammoInitialAmount);
					dict.GetInt("initialAltAmmoAmount", weapon.ammoAltInitialAmount);
					dict.GetInt("ammoConsumption", weapon.ammoConsumption);
					dict.GetBool("canSever", weapon.canSever);
					dict.GetBool("hasScope", weapon.hasScope);
					dict.GetBool("owned", weapon.startingWeapon);
					dict.GetBool("allowTracking", weapon.allowTracking);
					dict.GetBool("allowTrackingMulti", weapon.allowTrackingMulti);
					dict.GetFloat("maxTrackingDegrees", weapon.maxTrackingDegrees);
					dict.GetInt("projectileParticle", weapon.projectileParticleID);
					dict.GetInt("altProjectileParticle", weapon.projectileAltParticleID);
					dict.GetInt("projectileSnipe", weapon.projectileSnipeParticleID);
					dict.GetInt("altProjectileSnipe", weapon.projectileAltSnipeParticleID);
					weapons[i] = weapon;
				}
			}
			
			//load ammo info
			kDefManager defManager;
			defManager.LoadFile("defs/ammoInfo.txt");
			for (int i = 0; i < maxWeapons; i++)
			{
				kStr ammoName = weaponAmmoNames[i];
				kDictMem@ dict = defManager.GetEntry(ammoName);
				if (dict !is null)
				{
					int value;
					dict.GetInt("max", value, 0);
					weapons[i].ammoMax = value;
				}
				
				ammoName = weaponAltAmmoNames[i];
				@dict = defManager.GetEntry(ammoName);
				if (dict !is null)
				{
					int value;
					dict.GetInt("max", value, 0);
					weapons[i].ammoAltMax = value;
				}
			}
		}
		
		class Weapon
		{
			int actorID; //the kWeapon actor ID
			int pickupActorID;
			int charMeshIndex; //char attached weapon index
			int ammoInitialAmount;
			int ammoAltInitialAmount;
			int ammoConsumption;
			int ammoMax;
			int ammoAltMax;
			bool canSever;
			bool hasScope;
			bool startingWeapon;
			bool allowTracking;
			bool allowTrackingMulti;
			float maxTrackingDegrees;
			int projectileParticleID;
			int projectileAltParticleID;
			int projectileSnipeParticleID;
			int projectileAltSnipeParticleID;
			
			//------------------------------------------------------------------------------------------------------------------------
			Weapon()
			{
				
			}
			//------------------------------------------------------------------------------------------------------------------------
			Weapon@ opAssign(const Weapon &in other)
			{
				actorID = other.actorID;
				pickupActorID = other.pickupActorID;
				charMeshIndex = other.charMeshIndex;
				ammoInitialAmount = other.ammoInitialAmount;
				ammoAltInitialAmount = other.ammoAltInitialAmount;
				ammoConsumption = other.ammoConsumption;
				ammoMax = other.ammoMax;
				ammoAltMax = other.ammoAltMax;
				canSever = other.canSever;
				hasScope = other.hasScope;
				startingWeapon = other.startingWeapon;
				allowTracking = other.allowTracking;
				allowTrackingMulti = other.allowTrackingMulti;
				maxTrackingDegrees = other.maxTrackingDegrees;
				projectileParticleID = other.projectileParticleID;
				projectileAltParticleID = other.projectileAltParticleID;
				projectileSnipeParticleID = other.projectileSnipeParticleID;
				projectileAltSnipeParticleID = other.projectileAltSnipeParticleID;
				return this;
			}
			//------------------------------------------------------------------------------------------------------------------------
			kStr ToString()
			{
				return "actorID: " + actorID + " pickupActorID: " + pickupActorID + " charMeshIndex: " + charMeshIndex + " ammoInitialAmount: " + ammoInitialAmount +
							" ammoAltInitialAmount: " + ammoAltInitialAmount + " ammoConsumption: " + ammoConsumption + " ammoMax: " + ammoMax +
							" ammoAltMax: " + ammoAltMax + " canSever: " + canSever + " hasScope: " + hasScope + " startingWeapon: " + startingWeapon +
							"\nallowTracking: " + allowTracking + " allowTrackingMulti: " + allowTrackingMulti + " maxTrackingDegrees: " + maxTrackingDegrees +
							" projectileParticleID: " + projectileParticleID + " projectileAltParticleID: " + projectileAltParticleID + " projectileSnipeParticleID: " + projectileSnipeParticleID +
							" projectileAltSnipeParticleID: " + projectileAltSnipeParticleID;
			}
			//------------------------------------------------------------------------------------------------------------------------
		};
	}
}
