#define NETMSG_TIME 300
#define NETMSG_GRAVITY "Player Gravity set to " + BP::playerGravity
#define NETMSG_SPEED "Player Speed set to " + BP::playerSpeed
#define NETMSG_DAMAGE "Player Damage set to " + BP::playerDamage
#define NETMSG_RECEIVEDAMAGE "Player Receive Damage set to " + BP::playerReceiveDamage
#define NETMSG_DOUBLE_JUMP "Double Jump " + (BP::playerCanDoubleJump ? "Enabled" : "Disabled")
#define NETMSG_GROUND_DASH "Ground Dash " + (BP::playerCanGroundDash ? "Enabled" : "Disabled")
#define NETMSG_AIR_DASH "Air Dash " + (BP::playerCanAirDash ? "Enabled" : "Disabled")
#define NETMSG_GROUND_POUND "Ground Pound " + (BP::playerCanGroundPound ? "Enabled" : "Disabled")
#define NETMSG_SHOW_PLAYER_DAMAGE "Show Player Damage " + (BP::showPlayerDamage ? "Enabled" : "Disabled")
#define NETMSG_CLOAK "Player Cloaking " + (BP::playerCloaking ? "Enabled" : "Disabled")

void SendSound(const int soundID)
{
	NetWorkWrite32("void n()", soundID);
}

void slap(const int16 actorNetID)
{
	if (IsNetServer)
	{
		kActor@ actor = Network.GetNetActor(actorNetID);
		if (!BP::Actor::Exists(actor) || BP::Actor::IsDead(actor))
		{
			PrintLog("Can not slap player " + actorNetID);
			return;
		}
		int x = Math::RandMax(16);
		int y = Math::RandMax(16);
		int z = Math::RandMax(16);
		int power = Math::RandMax(16);
		int16 flags = (x << 0) | (y << 4) | (z << 8) | (power << 12);
		SlapPlayer(actorNetID, flags);
		PrintLog("Slapped player " + actorNetID);
		NetWorkWrite("void o()", actorNetID, flags);
	}
}

void SlapPlayer(int16 actorNetID, int16 flags)
{
	kActor@ actor = Network.GetNetActor(actorNetID);
	if (actor is null)
	{
		return;
	}
	
	if (actor.WorldComponent() !is null)
	{
		//not in water and has a movement component
		if ((actor.WorldComponent().Flags() & (WCF_UNDER_WATER | WCF_ON_WATER)) == 0 && actor.MovementComponent() !is null)
		{
			actor.MovementComponent().Flags() |= MCF_DISABLE_AIR_FRICTION_UNTIL_LAND;
		}
	}
	int16 x = ((flags & 0xf) >> 0) - 7;
	int16 y = ((flags & 0xf0) >> 4) - 7;
	int16 z = ((flags & 0xf00) >> 8) - 7;
	int16 power = 30 + ((flags & 0xf000) >> 12);	
	kVec3 vel = kVec3(x, y, float(z) / 2).Normalize();
	vel.x *= power;
	vel.y *= power;
	vel.z *= power * 0.8f;
	BP::slapVelocity = vel;
	actor.PlaySound(BP_SFX_Slap);
	if (actor.MovementComponent() !is null)
	{
		actor.MovementComponent().Velocity() = BP::slapVelocity;
	}
}

void showmods()
{
	PrintLog("----------Current Mods----------");
	PrintError(NETMSG_GRAVITY);
	PrintError(NETMSG_SPEED);
	PrintError(NETMSG_DAMAGE);
	PrintError(NETMSG_RECEIVEDAMAGE);
	PrintError(NETMSG_DOUBLE_JUMP);
	PrintError(NETMSG_GROUND_DASH);
	PrintError(NETMSG_AIR_DASH);
	PrintError(NETMSG_GROUND_POUND);
	PrintError(NETMSG_SHOW_PLAYER_DAMAGE);
	PrintError(NETMSG_CLOAK);
	PrintLog("");
}

void SetAllNetModsSilent()
{
	NetWorkWrite("void c()", BP::playerCanDoubleJump ? 1 : 0, 1);
	NetWorkWrite("void d()", BP::playerCanGroundDash ? 1 : 0, 1);
	NetWorkWrite("void e()", BP::playerCanAirDash ? 1 : 0, 1);
	NetWorkWrite("void f()", BP::playerCanGroundPound ? 1 : 0, 1);
	NetWorkWrite("void g()", int16(BP::playerReceiveDamage * 10.0f), 1);
	NetWorkWrite("void h()", int16(BP::playerDamage * 10.0f), 1);
	NetWorkWrite("void i()", int16(BP::playerSpeed * 100.0f), 1);
	NetWorkWrite("void j()", int16(BP::playerGravity * 100.0f), 1);
	NetWorkWrite("void k()", BP::showPlayerDamage ? 1 : 0, 1);
	NetWorkWrite("void l()", BP::playerCloaking ? 1 : 0, 1);
	SendMPGameMode(BP::mapScript.gameModeMP);
}

void SendMPGameMode(int gameMode)
{
	NetWorkWrite("void m()", gameMode);
}

void SetPlayerCloak(const bool cloak)
{
	if (IsMultiplayer())
	{
		BP::playerCloaking = cloak;
		SetCloakForAllPlayers(cloak);
		AddHudMessage(NETMSG_CLOAK);
		if (IsNetServer)
		{
			NetWorkWrite("void l()", cloak ? 1 : 0, 0);
		}
	}
}

void SetCloakForAllPlayers(const bool cloak)
{
	FOREACH_ACTOR
	{
		if (actor.Type() == kActor_Player)
		{
			if (cloak)
			{
				actor.RenderMeshComponent().Flags() |= RMCF_CLOAKED;
			}
			else
			{
				actor.RenderMeshComponent().Flags() &= ~RMCF_CLOAKED;
			}
		}
	}
}

void SetShowPlayerDamage(const bool showDamage)
{
	if (IsMultiplayer())
	{
		BP::showPlayerDamage = showDamage;
		AddHudMessage(NETMSG_SHOW_PLAYER_DAMAGE);
		if (IsNetServer)
		{
			NetWorkWrite("void k()", showDamage ? 1 : 0, 0);
		}
	}
}

void SetDoubleJump(const bool canDoubleJump)
{
	if (IsMultiplayer())
	{
		BP::playerCanDoubleJump = canDoubleJump;
		AddHudMessage(NETMSG_DOUBLE_JUMP);
		if (IsNetServer)
		{
			NetWorkWrite("void c()", canDoubleJump ? 1 : 0, 0);
		}
	}
}

void SetGroundDash(const bool canGroundDash)
{
	if (IsMultiplayer())
	{
		BP::playerCanGroundDash = canGroundDash;
		AddHudMessage(NETMSG_GROUND_DASH);
		if (IsNetServer)
		{
			NetWorkWrite("void d()", canGroundDash ? 1 : 0, 0);
		}
	}
}

void SetAirDash(const bool canAirDash)
{
	if (IsMultiplayer())
	{
		BP::playerCanAirDash = canAirDash;
		AddHudMessage(NETMSG_AIR_DASH);
		if (IsNetServer)
		{
			NetWorkWrite("void e()", canAirDash ? 1 : 0, 0);
		}
	}
}

void SetGroundPound(const bool canGroundPound)
{
	if (IsMultiplayer())
	{
		BP::playerCanGroundPound = canGroundPound;
		AddHudMessage(NETMSG_GROUND_POUND);
		if (IsNetServer)
		{
			NetWorkWrite("void f()", canGroundPound ? 1 : 0, 0);
		}
	}
}

void SetDamage(const float damage)
{
	if (IsMultiplayer())
	{
		BP::playerDamage = damage;
		AddHudMessage(NETMSG_DAMAGE);
		if (IsNetServer)
		{
			NetWorkWrite("void g()", int16(damage * 10.0f), 0);
		}
	}
}

void SetReceiveDamage(const float damage)
{
	if (IsMultiplayer())
	{
		BP::playerReceiveDamage = damage;
		AddHudMessage(NETMSG_RECEIVEDAMAGE);
		if (IsNetServer)
		{
			NetWorkWrite("void h()", int16(damage * 10.0f), 0);
		}
	}
}

void SetSpeed(const float speed)
{
	if (IsMultiplayer())
	{
		BP::playerSpeed = speed;
		SetSpeedForAllActors(speed);
		AddHudMessage(NETMSG_SPEED);
		if (IsNetServer)
		{
			NetWorkWrite("void i()", int16(speed * 100.0f), 0);
		}
	}
}

void SetSpeedForAllActors(const float speed)
{
	FOREACH_ACTOR
	{
		if (actor.Type() == kActor_Player)
		{
			BP::Actor::SetSpeed(@actor, speed);
		}
	}
}

void SetGravity(const float gravity)
{
	if (IsMultiplayer())
	{
		BP::playerGravity = gravity;
		SetGravityForAllPlayers(gravity);
		AddHudMessage(NETMSG_GRAVITY);
		if (IsNetServer)
		{
			NetWorkWrite("void j()", int16(gravity * 100.0f), 0);
		}
	}
}

void SetGravityForAllPlayers(const float gravity)
{
	FOREACH_ACTOR
	{
		if (actor.Type() == kActor_Player)
		{
			BP::Actor::SetGravity(@actor, gravity);
		}
	}
}

void NetWorkWriteActor(kActor@ actor)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader("void b()");
	NetSend::WriteInt32(actor.TID());
	Network.TransmitToAll();
}

void NetWorkWriteActor(kActor@ actor, array<int>& int32Args, array<int16>& int16Args)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader("void a()");
	NetSend::WriteInt32(actor.TID());
	NetSend::WriteInt16(int16(int32Args.length()));
	NetSend::WriteInt16(int16(int16Args.length()));
	for (int i = 0; i < int(int32Args.length()); i++)
	{
		NetSend::WriteInt32(int32Args[i]);
	}
	for (int i = 0; i < int(int16Args.length()); i++)
	{
		NetSend::WriteInt16(int16Args[i]);
	}
	Network.TransmitToAll();
}

void NetWorkWriteActorInt16(kActor@ actor, array<int16>& int16Args)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader("void q()");
	NetSend::WriteInt32(actor.TID());
	NetSend::WriteInt16(int16(int16Args.length()));
	for (int i = 0; i < int(int16Args.length()); i++)
	{
		NetSend::WriteInt16(int16Args[i]);
	}
	Network.TransmitToAll();
}

void NetWorkWriteActorInt32(kActor@ actor, array<int>& int32Args)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader("void p()");
	NetSend::WriteInt32(actor.TID());
	NetSend::WriteInt16(int16(int32Args.length()));
	for (int i = 0; i < int(int32Args.length()); i++)
	{
		NetSend::WriteInt32(int32Args[i]);
	}
	Network.TransmitToAll();
}

void NetWorkWrite(const kStr&in header)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader(header);
	Network.TransmitToAll();
}

void NetWorkWrite(const kStr&in header, const int16 data)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader(header);
	NetSend::WriteInt16(data);
	Network.TransmitToAll();
}

void NetWorkWrite(const kStr&in header, const int16 data, const int16 data2)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader(header);
	NetSend::WriteInt16(data);
	NetSend::WriteInt16(data2);
	Network.TransmitToAll();
}

void NetWorkWrite32(const kStr&in header, const int data)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader(header);
	NetSend::WriteInt32(data);
	Network.TransmitToAll();
}

void NetWorkWrite32(const kStr&in header, const int data, const int data2)
{
	NetSend::Clear();
	NetSend::WriteScriptHeader(header);
	NetSend::WriteInt32(data);
	NetSend::WriteInt32(data2);
	Network.TransmitToAll();
}

void AddHudMessage(const kStr&in msg, const int flags = 0)
{
	bool silent = (flags & (1 << 0)) != 0;
	if (!silent)
	{
		Hud.AddMessage(msg, NETMSG_TIME);
	}
}

//-----Network callback functions-----
void i() // Player Speed Change
{
    if (IsNetClient)
    {
		BP::playerSpeed = (NetRecv::ReadInt16() / 100.0f); 
		SetSpeedForAllActors(BP::playerSpeed);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_SPEED, flags);
	}
}

void j() // Gravity Changed
{
    if (IsNetClient)
    {
		BP::playerGravity = (NetRecv::ReadInt16() / 100.0f); 
		SetGravityForAllPlayers(BP::playerGravity);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_GRAVITY, flags);
	}
}

void g() // Damage Changed
{
    if (IsNetClient)
    {
		BP::playerDamage = (NetRecv::ReadInt16() / 10.0f); 
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_DAMAGE, flags);
	}
}

void h() // Defense Change
{
    if (IsNetClient)
    {
		BP::playerReceiveDamage = (NetRecv::ReadInt16() / 10.0f); 
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_RECEIVEDAMAGE, flags);
	}
}

void c() //Double Jump Changed
{
    if (IsNetClient)
    {
		BP::playerCanDoubleJump = (NetRecv::ReadInt16() == 1);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_DOUBLE_JUMP, flags);
	}
}

void d() //Ground Dash Changed
{
    if (IsNetClient)
    {
		BP::playerCanGroundDash = (NetRecv::ReadInt16() == 1);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_GROUND_DASH, flags);
	}
}

void e() //Air Dash Changed
{
    if (IsNetClient)
    {
		BP::playerCanAirDash = (NetRecv::ReadInt16() == 1);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_AIR_DASH, flags);
	}
}

void f() // Ground Pound Changed
{
    if (IsNetClient)
    {
		BP::playerCanGroundPound = (NetRecv::ReadInt16() == 1);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_GROUND_POUND, flags);
	}
}

void k() // Show Player Damage Changed
{
    if (IsNetClient)
    {
		BP::showPlayerDamage = (NetRecv::ReadInt16() == 1);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_SHOW_PLAYER_DAMAGE, flags);
	}
}

void l() // Player Cloaking Changed
{
    if (IsNetClient)
    {
		BP::playerCloaking = (NetRecv::ReadInt16() == 1);
		SetCloakForAllPlayers(BP::playerCloaking);
		int16 flags = NetRecv::ReadInt16();
		AddHudMessage(NETMSG_CLOAK, flags);
	}
}

void o() //Slap player
{
    if (IsNetClient)
    {
		int16 actorNetID = NetRecv::ReadInt16();
		int16 flags = NetRecv::ReadInt16();
		SlapPlayer(actorNetID, flags);
	}
}

void m() //MP Game mode Changed
{
    if (IsNetClient)
    {
		BP::mapScript.gameModeMP = NetRecv::ReadInt16();
	}
}

void n() //Received Sound to play
{
    if (IsNetClient)
    {
		BP::Audio::PlaySoundOnNonHiddenActor(NetRecv::ReadInt32());
	}
}

void a() //Network Receive Actor Data (int32s and int16s)
{
    if (IsNetClient)
    {
		int tid = NetRecv::ReadInt32();
		kActor @actor = BP::Actor::FindActorWithTID(tid);
		if (actor !is null)
		{
			int int32ArgLength = NetRecv::ReadInt16();
			int int16ArgLength = NetRecv::ReadInt16();
			array<int> int32Args();
			array<int16> int16Args();
			for (int i = 0; i < int32ArgLength; i++)
			{
				int32Args.insertLast(NetRecv::ReadInt32());
			}
			for (int i = 0; i < int16ArgLength; i++)
			{
				int16Args.insertLast(NetRecv::ReadInt16());
			}
			BP::UserEvent::ActorNetReceive(@actor, int32Args, int16Args);
		}
	}
}

void b() //Network Receive Actor Data (no args)
{
    if (IsNetClient)
    {
		int tid = NetRecv::ReadInt32();
		kActor @actor = BP::Actor::FindActorWithTID(tid);
		if (actor !is null)
		{
			BP::UserEvent::ActorNetReceive(@actor, array<int> = {}, array<int16> = {});
		}
	}
}

void p() //Network Receive Actor Data (int32s only)
{
    if (IsNetClient)
    {
		int tid = NetRecv::ReadInt32();
		kActor @actor = BP::Actor::FindActorWithTID(tid);
		if (actor !is null)
		{
			int int32ArgLength = NetRecv::ReadInt16();
			array<int> int32Args();
			for (int i = 0; i < int32ArgLength; i++)
			{
				int32Args.insertLast(NetRecv::ReadInt32());
			}
			BP::UserEvent::ActorNetReceive(@actor, int32Args, array<int16> = {});
		}
	}
}

void q() //Network Receive Actor Data (int16s only)
{
    if (IsNetClient)
    {
		int tid = NetRecv::ReadInt32();
		kActor @actor = BP::Actor::FindActorWithTID(tid);
		if (actor !is null)
		{
			int int16ArgLength = NetRecv::ReadInt16();
			array<int16> int16Args();
			for (int i = 0; i < int16ArgLength; i++)
			{
				int16Args.insertLast(NetRecv::ReadInt16());
			}
			BP::UserEvent::ActorNetReceive(@actor, array<int> = {}, int16Args);
		}
	}
}
