namespace BP
{	
	namespace Actor
	{
		kActor@ Null = null; //for setting SetTarget() to null
		
		//------------------------------------------------------------------------------------------------------------------------
		int NetID(kActor@ actor)
		{
			int netID = actor.NetID();
			if (netID < 0)
			{
				netID = 0;
			}
			return netID;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetSpeed(kActor@ actor, const float speed)
		{
			actor.MovementComponent().Friction() = speed;
			actor.MovementComponent().AirFriction() = speed;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetGravity(kActor@ actor, const float gravity)
		{
			actor.MovementComponent().Gravity() = gravity;
		}
		//------------------------------------------------------------------------------------------------------------------------
		kActor@ FindActorWithTID(const int tagID)
		{
			FOREACH_ACTOR
			{
				if (actor.TID() == tagID)
				{
					return @actor;
				}
			}
			
			return null;
		}
		//------------------------------------------------------------------------------------------------------------------------
		array<kActor@> FindActorsWithTID(const int tagID)
		{
			array<kActor@> result;
			FOREACH_ACTOR
			{
				if (actor.TID() == tagID)
				{
					result.insertLast(@actor);
				}
			}
			
			return result;
		}
		//------------------------------------------------------------------------------------------------------------------------
		// returns the first actor that have the same Type
		//------------------------------------------------------------------------------------------------------------------------
		kActor@ FindActorWithType(const int typeID)
		{
			array<kActor@> result;
			FOREACH_ACTOR
			{
				if (actor.Type() == typeID)
				{
					return @actor;
				}
			}
			
			return null;
		}
		//------------------------------------------------------------------------------------------------------------------------
		// returns a list of all the actors in that have the same Type
		//------------------------------------------------------------------------------------------------------------------------
		array<kActor@> FindActorsWithType(const int typeID)
		{
			array<kActor@> result;
			FOREACH_ACTOR
			{
				if (actor.Type() == typeID)
				{
					result.insertLast(@actor);
				}
			}
			return result;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void AIMoveToTarget(kActor@ actor, kActor @target)
		{
			DebugAssert(actor !is null, "[BP::Actor::AIMoveToTarget] actor is null");
			DebugAssert(actor.EnemyAIComponent() !is null, "[BP::Actor::AIMoveToTarget] actor doesn't have EnemyAIComponent");
			actor.EnemyAIComponent().PathTarget().SetTarget(@target);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void AIAttackTarget(kActor@ actor, kActor @target)
		{
			DebugAssert(actor !is null, "[BP::Actor::AIAttackTarget] actor is null");
			DebugAssert(actor.EnemyAIComponent() !is null, "[BP::Actor::AIAttackTarget] actor doesn't have EnemyAIComponent");
			actor.EnemyAIComponent().SightTarget().SetTarget(@target);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void AIIgnorePlayer(kActor@ actor, bool ignore)
		{
			DebugAssert(actor !is null, "[BP::Actor::AIIgnorePlayer] actor is null");
			DebugAssert(actor.EnemyAIComponent() !is null, "[BP::Actor::AIIgnorePlayer] actor doesn't have EnemyAIComponent");
			actor.EnemyAIComponent().SetNoUpdateTarget(ignore, ignore);
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns true if actor is on the ground
		//------------------------------------------------------------------------------------------------------------------------
		bool IsGrounded(kActor@ actor)
		{
			DebugAssert(actor !is null, "[BP::Actor::IsGrounded] actor is null");
			DebugAssert(actor.MovementComponent() !is null, "[BP::Actor::IsGrounded] actor doesn't have a movement component");
			return (actor.MovementComponent().Flags() & MCF_FALLING) == 0;
			
			//Also works (using WorldComponent)
			// DebugAssert(actor !is null, "[BP::Actor::IsGrounded] actor is null");
			// DebugAssert(actor.WorldComponent() !is null, "[BP::Actor::IsGrounded] actor doesn't have a world component");
			// return Math::Approximately(actor.Origin().z, actor.WorldComponent().FloorHeight());
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns the forward direction of the actor
		//------------------------------------------------------------------------------------------------------------------------
		kVec3 Forward(kActor@ actor)
		{
			return kVec3(0.0f, 1.0f, 0.0f) * kQuat(actor.Pitch(), 0.0f, actor.Yaw());
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns the right direction of the actor
		//------------------------------------------------------------------------------------------------------------------------
		kVec3 Right(kActor@ actor)
		{
			return kVec3(1.0f, 0.0f, 0.0f) * kQuat(actor.Pitch(), 0.0f, actor.Yaw());
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns the up direction of the actor
		//------------------------------------------------------------------------------------------------------------------------
		kVec3 Up(kActor@ actor)
		{
			return kVec3(0.0f, 0.0f, 1.0f) * kQuat(actor.Pitch(), 0.0f, actor.Yaw());
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns the up direction of the actor
		//------------------------------------------------------------------------------------------------------------------------
		kVec3 Direction(kActor@ actor, float x, float y, float z)
		{
			return kVec3(x, y, z) * kQuat(actor.Pitch(), 0.0f, actor.Yaw());
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns the center position of the actor
		//------------------------------------------------------------------------------------------------------------------------
		kVec3 CenterOrigin(kActor @actor, float heightRatio = 0.5f)
		{
			return kVec3(actor.Origin().x, actor.Origin().y, actor.Origin().z + (actor.WorldComponent().Height() * heightRatio) + actor.WorldComponent().HeightOffset());
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool InRadius(kActor @actor, kActor @targetActor, const float radius)
		{
			return actor.Origin().DistanceSq(targetActor.Origin()) < radius;
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool InRadius(kActor @actor, const kVec3& in pos, const float radius)
		{
			return actor.Origin().DistanceSq(pos) < radius;
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool InRadius(const kVec3& in point, const kVec3& in pos, const float radius)
		{
			return point.DistanceSq(pos) < radius;
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns true if the actor is not stale, not dead, not Hidden
		//------------------------------------------------------------------------------------------------------------------------
		bool IsActive(kActor @actor)
		{
			return Exists(actor) && !IsDead(@actor) && !IsHidden(@actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns true if actor is not null and not stale
		//------------------------------------------------------------------------------------------------------------------------
		bool Exists(kActor @actor)
		{
			return actor !is null && !actor.IsStale();
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool IsDead(kActor @actor)
		{
			return (actor.Flags() & AF_DEAD) != 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool IsHidden(kActor @actor)
		{
			return (actor.Flags() & AF_HIDDEN) != 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Returns true if not hidden or not dead (When there's a client to the player actor and they're not spectating (no way to tell the difference))
		//------------------------------------------------------------------------------------------------------------------------
		bool IsMPPlayerInGame(kActor @actor)
		{
			return !IsHidden(@actor) || !IsDead(@actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		// shows the text above this actor if they have a world component other wise will be displayed at the actors origin.
		//------------------------------------------------------------------------------------------------------------------------
		void ShowWorldText(kActor @srcActor, kActor @targetActor, kStr&in text, kVec3& in posOffset = Math::vecZero)
		{
			float charSpacing = 35.0f;
			
			kVec3 origin = targetActor.Origin();
			kWorldComponent @worldComponent = targetActor.WorldComponent();
			if (worldComponent !is null)
			{
				float radius = targetActor.WorldComponent().Radius() * 0.75f;
				posOffset.x += Math::RandRange(-radius, radius);
				posOffset.y += Math::RandRange(-radius, radius);
				posOffset.z += (worldComponent.Height() + worldComponent.HeightOffset()) * 1.75f;
			}
			
			int textLength = BP::String::Length(text);
			for (int i = 0; i < textLength; ++i)
			{
				kActor @charController = BP::Spawn::Actor(BP_Actor_WorldChar, origin);
				DebugAssert(charController !is null, "[BP::Actor::ShowWorldText] charController is null");
				float offset = (-((textLength - 1) / 2.0f) + i) * charSpacing;
				BP::WorldChar@ charScript = BP::UserEvent::GetCharScript(@charController);
				DebugAssert(charScript !is null, "[BP::Actor::ShowWorldText] charScript is null");
				charScript.Setup(@srcActor, @targetActor, text[i], origin, posOffset, offset);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetAIPath(const int tagID, const array<int>& pathPoints, const bool ignorePlayer)
		{
			kActor @actor = BP::Actor::FindActorWithTID(tagID);
			SetAIPath(@actor, pathPoints, ignorePlayer);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetAIPath(kActor@ actor, const array<int>& pathPoints, const bool ignorePlayer)
		{
			if (actor !is null)
			{
				BP::AI @ai = BP::UserEvent::GetAIScript(@actor);
				DebugAssert(ai !is null, "[BP::Actor::SetAIPath] actor does not have a BP::AI script");
				ai.SetPath(pathPoints, ignorePlayer);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void GenerateFlags(uint flags1, uint flags2)
		{
			if ((flags1 & GENERATE_FLAG_HEALTH2) != 0)
			{
				//spawn health 2
			}
			
			if ((flags2 & GENERATE_FLAG2_NUKE) != 0)
			{
				//spawn nuke ammo
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	}
}
